/*   
 * Copyright (C) 2006 Daniel Fernandez <ergot86@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include <stdio.h>
#include <errno.h>
#include "dbg.h"

void DSnap(struct Debug *dbg)
{
	wait(NULL) ;
	ptrace(PTRACE_GETSIGINFO,dbg->pid,0,&dbg->sig) ;
	ptrace(PTRACE_GETFPXREGS,dbg->pid,0,&dbg->fregs) ;
	ptrace(PTRACE_GETREGS,dbg->pid,0,&dbg->regs) ;
	dbg->u_debugreg[0] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[0])) ;
	dbg->u_debugreg[1] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[1])) ;
	dbg->u_debugreg[2] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[2])) ;
	dbg->u_debugreg[3] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[3])) ;
	dbg->u_debugreg[4] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[4])) ;
	dbg->u_debugreg[5] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[5])) ;
	dbg->u_debugreg[6] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[6])) ;
	dbg->u_debugreg[7] = ptrace(PTRACE_PEEKUSR,dbg->pid,offsetof(struct user,u_debugreg[7])) ;

	if (dbg->DbgH)
		dbg->DbgH(dbg) ;
}


void SetRegs(struct Debug *dbg)
{
	ptrace(PTRACE_SETFPXREGS,dbg->pid,0,&dbg->fregs) ;
	ptrace(PTRACE_SETREGS,dbg->pid,0,&dbg->regs) ;
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[0]),dbg->u_debugreg[0]) ;
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[1]),dbg->u_debugreg[1]) ;
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[2]),dbg->u_debugreg[2]) ;
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[3]),dbg->u_debugreg[3]) ;
/*
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[4]),dbg->u_debugreg[4]) ;
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[5]),dbg->u_debugreg[5]) ;
*/
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[6]),dbg->u_debugreg[6]) ;
	ptrace(PTRACE_POKEUSR,dbg->pid,offsetof(struct user,u_debugreg[7]),dbg->u_debugreg[7]) ;
}


int Trace(struct Debug *dbg, unsigned long fail_action)
{
	char path[21] ;
  
	sprintf(path,"/proc/%d/mem",dbg->pid) ;
	dbg->mem = open(path,O_RDWR) ;
  
	if (dbg->mem == -1)
	{
		ptrace(fail_action,dbg->pid) ;
		return -1 ;
	}

	dbg->sysemu = FALSE ;
	dbg->bps = NULL ;
	dbg->wps = NULL ;
	dbg->DbgH = NULL ;
	DSnap(dbg) ;
	ptrace(PTRACE_SETOPTIONS,dbg->pid,0,PTRACE_O_ALL) ;
	return 0 ;
}

  
int TraceFile(const char *file, char *const argv[], char *const env[], struct Debug *dbg)
{
	dbg->pid = fork() ;
  
	switch(dbg->pid)
	{
		case 0:
			if (ptrace(PTRACE_TRACEME))
				return -1 ;

			execve(file,argv,env) ;
			
		case -1:
			return -1 ;
	}

	return Trace(dbg,PTRACE_KILL) ;
}


int Attach(struct Debug *dbg)
{
	return (ptrace(PTRACE_ATTACH,dbg->pid)) ? -1 : Trace(dbg,PTRACE_DETACH) ;
}


/* Posiblemente quieras llamar a RemoveAllBp() y RemoveAllWp() antes de Detach() */
int Detach(struct Debug *dbg)
{
	close(dbg->mem) ;
	return (ptrace(PTRACE_DETACH,dbg->pid,0,dbg->sig.si_signo))? -1 : 0 ;
}


int Kill(struct Debug *dbg)
{
	close(dbg->mem) ;
	return (ptrace(PTRACE_KILL,dbg->pid))? -1 : 0 ;
}

int SendSignal(struct Debug *dbg)
{
	int err ;

	ptrace(PTRACE_SETSIGINFO,dbg->pid,0,&dbg->sig) ;
	err = ptrace(PTRACE_CONT,dbg->pid,0,dbg->sig.si_signo) ;

	if (likely(!err))
		DSnap(dbg) ;

	return err ;
}

/* Bueno, lo que sigue es feo como pegarle a dios... */
#ifdef __i386__
#define FMAP "%lx-%lx %c%c%c%c %lx %02x:%02x %lu%[^\n]"
#endif

#ifdef __x86_64__
#define FMAP "%llx-%llx %c%c%c%c %llx %02x:%02x %lu%[^\n]"
#endif

int GetMapInfo(unsigned long addr,struct vm_map *map,struct Debug *dbg)
{
	unsigned int mi, ma ;
	char path[22], buf[PATH_MAX+25+sizeof(void*)*6], r, w, x, s ;
	FILE *f ;

	sprintf(path,"/proc/%d/maps",dbg->pid) ;
	f = fopen(path,"r") ;

	while(fscanf(f,FMAP,&map->start,&map->end,&r,&w,&x,&s,&map->offset,&ma,&mi,&map->inode,buf) != EOF)
	{
		if (addr >= map->start && addr <= map->end)
		{
			map->flags = 0 ;
			map->flags |= (r == 'r') ? VM_READ : 0 ;
			map->flags |= (w == 'w') ? VM_WRITE : 0 ;
			map->flags |= (x == 'x') ? VM_EXEC : 0 ;
			map->flags |= (s == 's') ? VM_SHARED : 0;
			map->dev = (ma<<8) | mi ;
			strcpy(map->file,strchr(buf,' ')) ;
			return 0 ;
		}
	}

	return -1 ;
}
/* ...Algun voluntario para realizarle un extreme makeover a la funcion anterior? */

