/*   
 * Copyright (C) 2006 Daniel Fernandez <ergot86@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */
#include <stdio.h>
#include <errno.h>
#include <asm/system.h>
#include "dbg.h"

int StepIn(struct Debug *dbg)
{
	//ptrace((dbg->sysemu)?PTRACE_SYSEMU_SINGLESTEP:PTRACE_SINGLESTEP,dbg->pid,0,0) ;
	struct BreakPoint *bps = NULL ;
	struct WatchPoint *wps = NULL ;
	struct ptrace_mprotect_args args ;
	int err ;

	if (!CheckWp(dbg))
	{
		wps = FindWp((unsigned long)dbg->sig.si_addr,dbg) ;
		args.addr = wps->addr ;
		args.len = PAGE_SIZE ;
		args.prot = wps->prot ;
		
		if (ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) == -1)
			return -1 ;
	}

	if (!CheckBp(dbg))
	{
		bps = FindBp(REG_IP,dbg) ;
		WriteMem(bps->addr,1,&bps->op,dbg) ;
	}

	err = ptrace(PTRACE_SINGLESTEP,dbg->pid,0,0) ;

	if (likely(!err))
		DSnap(dbg) ;

	if (wps)
	{
		args.prot = PROT_NONE ;
		ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) ;
		err |= 1 ;
	}

	if (bps)
	{
		WriteMem(bps->addr,1,&bps->bp,dbg) ;
		err |= 2 ;
	}

	return err ;
}


int StepOver(struct Debug *dbg)
{
	struct BreakPoint *bps = NULL ;
	struct WatchPoint *wps = NULL ;
	struct ptrace_mprotect_args args ;
	unsigned long addr, ret, x ;
	int err ;
	char op[15] ;

	addr = REG_IP ;

	if (ReadMem(addr,15,op,dbg) < 1)
		return -1 ;

	if (!CheckWp(dbg))
	{
		wps = FindWp((unsigned long)dbg->sig.si_addr,dbg) ;
		args.addr = wps->addr ;
		args.len = PAGE_SIZE ;
		args.prot = wps->prot ;
		
		if (ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) == -1)
			return -1 ;
	}

	if (!CheckBp(dbg))
	{
		bps = FindBp(REG_IP,dbg) ;
		WriteMem(bps->addr,1,&bps->op,dbg) ;
	}
	
	for(x=0;x<6;x++)
	{
		switch(op[x])
		{
		/*
			Los opcodes de las instrucciones inc y dec (en su forma de un solo byte) fueron reasignados
			al prefijo REX. Este prefijo solo es usado en submodo de 64-bits. Para evitar problemas
			en otros modos se deber compilar una version exclusiva de 32 bits.
		*/
#ifdef __x86_64__
			case REX:
			case REXB:
			case REXX:
			case REXXB:
			case REXR:
			case REXRB:
			case REXRX:
			case REXRXB:
			case REXW:
			case REXWB:
			case REXWX:
			case REXWXB:
			case REXWR:
			case REXWRB:
			case REXWRX:
			case REXWRXB:
#endif
			case PREFIX_CS:
			case PREFIX_DS:
			case PREFIX_ES:
			case PREFIX_FS:
			case PREFIX_GS:
			case PREFIX_SS:
			case OPERS:
			case ADDRS:
			case LOCK:
				break ;
			case REPNZ:
			case REPZ:
				
				while(addr == REG_IP)
				{
					if (err = ptrace(PTRACE_SINGLESTEP,dbg->pid,0,0))
						break ;

					DSnap(dbg) ;
				}
				
				goto out ;
				
			case CALL:
				
				if (op[x+1] & '\x30')
				{
					
			case CALL_NI:
#ifdef __i386__
			/* En submodo de 64-bits no se pueden realizar llamadas FAR directas */
			case CALL_FD:
#endif
					if (err = ptrace(PTRACE_SINGLESTEP,dbg->pid,0,0))
						goto out ;

					DSnap(dbg) ;
					ret = ptrace(PTRACE_PEEKDATA,dbg->pid,REG_SP) ;
					
					/* Puede que el breakpoint ya exista en ese caso invalidamos la direccion para no quitarlo */
					if (SetBp(ret,dbg) == -1)
						ret = -1 ;

					/* Ojo!, si BpOpcode se toma como una instruccion de mas de un byte todo hace krakatoa */
					while(REG_IP-1 != ret)
					{
						err |= (!CheckWp(dbg)) ? 1 : 0 ;
						err |= (!CheckBp(dbg)) ? 2 : 0 ;
						
						/* No usamos Run() para permitir hooks en syscalls */
						if (err || (err = Syscall(dbg)))
							break ;
						
					}
					
					REG_IP = ret ;
					SetRegs(dbg) ;
					RemoveBp(ret,dbg) ;
					goto _out ;
				}
				
			default:
				err = ptrace(PTRACE_SINGLESTEP,dbg->pid,0,0) ;
				goto out ;
		}
	}

out:
	if (likely(!err))
		DSnap(dbg) ;

_out:		
	if (wps)
	{
		args.prot = PROT_NONE ;
		ptrace(PTRACE_MPROTECT,dbg->pid,0,&args) ;
		err |= 1 ;
	}

	if (bps)
	{
		WriteMem(bps->addr,1,&bps->bp,dbg) ;
		err |= 2 ;
	}
	
	return err ;
}


int Run(struct Debug *dbg)
{
	int err = ptrace(PTRACE_CONT,dbg->pid,0,0) ;

	DSnap(dbg) ;
	return err ;
}


int Syscall(struct Debug *dbg)
{
	// ptrace((dbg->sysemu)?PTRACE_SYSEMU:PTRACE_SYSCALL,dbg->pid,0,0) ;
	int err = ptrace(PTRACE_SYSCALL,dbg->pid,0,0) ;

	DSnap(dbg) ;
	return err ;
}


/*****************************
 *                           *
 *        BREAKPOINTS        *
 *                           *
 *****************************/

/*
   Tenemos una variable global con el opcode escrito sobre la instruccion y una funcion
   para cambiarlo (retorna el anterior).

   Puede ser util para dificultar la deteccion de nuestros breakpoints, reemplazando la int 3
   por alguna instruccion de un byte que genere una excepcion (ojo! posiblemente no llegue un
   SIGTRAP sino un SIGILL).

   Una solucion mas "limpia" pero menos performante es utilizar watchpoints.
   
*/

char BpOpcode = INT_3 ;


/*
  Retorna un puntero a la estructura del breakpoint correspondiente a la direccion addr,
  o NULL si este no existe. la funcion no comprueba la integridad del breakpoint.
*/
struct BreakPoint * FindBp(unsigned long addr,struct Debug *dbg)
{
	struct BreakPoint *bps = dbg->bps ;
  
	while(bps)
	{
		if (bps->addr == addr)
			return bps ;
    
		bps = bps->next ;
	}

	return NULL ;
}
      
/*
  Un breakpoint puede ser "pisado". Esta funcion retorna TRUE si el breakpoint
  sigue intacto o FALSE en caso contrario.
*/ 
int StillBpSet(struct BreakPoint *bps,struct Debug *dbg)
{
	char op ;
  
	ReadMem(bps->addr,1,&op,dbg) ;
  
	if (bps->bp != op)
		return FALSE ;

	return TRUE ;
}

/*
  Esta funcion verifica que el puntero de instruccion se encuentre en la lista
  de breakpoints y que el breakpoint correspondiente este integro. Si ambas
  condiciones se cumple la funcion retorna 0.
  
  Si el breakpoint fue encontrado en la estructura pero fue "pisado" la funcion
  retorna 1.

  NOTA: aunque la instruccion sea int 03, si el breakpoint no fue colocado con
  SetBp(), la funcion retorna -1 (y esta bien que sea de esta manera).
*/
  
int CheckBp(struct Debug *dbg)
{
	struct BreakPoint *bps = FindBp(REG_IP,dbg) ;

	if (!bps)
		return -1 ;

	return (StillBpSet(bps,dbg) == TRUE)? 0 : 1 ;
}

/*
   retorno: -1 error (o si ya existe), 0 exito.

   TODO: si esto se vuelve muy lento para gran cantidad de breakpoints,
   implementar un cache de la cola de la lista.
*/
int SetBp(unsigned long addr,struct Debug *dbg)
{
	struct BreakPoint *bpc = NULL ;
	struct BreakPoint *bps = dbg->bps ;

	while(bps)
	{
		if (bps->addr == addr)
			return -1 ;
	
		bpc = bps ;
		bps = bps->next ;
	}

	if (!bpc)
		bps = dbg->bps = malloc(sizeof(struct BreakPoint)) ;
	else
		bps = bpc->next = malloc(sizeof(struct BreakPoint)) ;

	if (!bps)
		return -1 ;

	if (ReadMem(addr,1,&bps->op,dbg) != 1 || WriteMem(addr,1,&BpOpcode,dbg) != 1)
	{
		free(bps) ;
		
		if (!bpc)
			dbg->bps = NULL ;
		else
			bpc->next = NULL ;
    
		return -1 ;
	}
	
	bps->next = NULL ;
	bps->addr = addr ;
	bps->bp = BpOpcode ;
	return 0 ;
}


int RemoveBp(unsigned long addr,struct Debug *dbg)
{
	struct BreakPoint *bpc = NULL ;
	struct BreakPoint *bps = dbg->bps ;

	while(bps)
	{
		if (bps->addr == addr)
		{
			WriteMem(addr,1,&bps->op,dbg) ;
			
			if (!bpc)
				dbg->bps = bps->next ;
			else
				bpc->next = bps->next ;
			
			free(bps) ;
			return 0 ;
		}
       
		bpc = bps ;
		bps = bps->next ;
	}
	
	return -1 ;
}


unsigned int RemoveAllBp(struct Debug *dbg)
{
	struct BreakPoint *bpc = NULL ;
	struct BreakPoint *bps = dbg->bps ;
	unsigned int x = 0 ;
  
	while(bps)
	{
		bpc = bps ;
		bps = bps->next ;
		WriteMem(bpc->addr,1,&bpc->op,dbg) ;
		free(bpc) ;
		x++ ;
	}
  
	dbg->bps = NULL ;
	return x ;
}


int ResetBp(unsigned long addr,struct Debug *dbg)
{
	struct BreakPoint *bps = FindBp(addr,dbg) ;
  
	if (!bps)
		return -1 ;

	if (StillBpSet(bps,dbg) == FALSE)
	{
		ReadMem(addr,1,&bps->op,dbg) ;
		WriteMem(addr,1,&bps->bp,dbg) ;
	}
	
	return 0 ;
}


unsigned int ResetAllBp(struct Debug *dbg)
{
	struct BreakPoint *bps = dbg->bps ;
	unsigned int x = 0 ;
  
	while(bps)
	{
		if (StillBpSet(bps,dbg) == FALSE)
		{
			ReadMem(bps->addr,1,&bps->op,dbg) ;
			WriteMem(bps->addr,1,&bps->bp,dbg) ;
		}
		
		bps = bps->next ;
		x++ ;
	}

	return x ;
}

