/*   
 * Copyright (C) 2006 Daniel Fernandez <ergot86@gmail.com>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
		 
#include <stdio.h>
#define _ISOC99_SOURCE
#include <stdlib.h>
#include <string.h>
#include <linux/limits.h>
#include <libdbg/dbg.h>
#define JIM_EMBEDDED
#include "jim.h"

#define MAX_CMD 256

struct commands
{
	char name[17] ;
	int (*dbgCommand) (Jim_Interp *interp, int argc, Jim_Obj *const *argv) ;

} Commands[MAX_CMD] ;

unsigned int CmdIndex = 0 ;
unsigned long dbgIndex = 0 ;


int dbgAddCommand(const char *name, int (*dbgCommand) (Jim_Interp *interp, int argc, Jim_Obj *const *argv))
{
	if (CmdIndex == MAX_CMD)
		return -1 ;
 
	strncpy(Commands[CmdIndex].name, name, 16) ;
	Commands[CmdIndex].name[16] = '\0' ;
	Commands[CmdIndex].dbgCommand = dbgCommand ;
	CmdIndex++ ;
	return 0 ;
}


static inline void SetVar(void *addr, void *val, size_t size)
{
	switch(size)
	{
		case sizeof(char):
			*((char *) addr) = *((char *) val) ;
			break ;
		case sizeof(short):
			*((short *) addr) = *((short *) val) ;
			break ;
		case sizeof(int):
			*((int *) addr) = *((int *) val) ;
			break ;
		case sizeof(long):
			*((long *) addr) = *((long *) val) ;
			break ;
		case (sizeof(long)*2):
			((long *) addr)[0] = ((long *) val)[0] ;
			((long *) addr)[1] = ((long *) val)[1] ;
	}
}


static inline void fVar(void *addr, char *buf, size_t size)
{
	switch(size)
	{
		case sizeof(char):
			sprintf(buf, "0x%02hhx", *((char *) addr)) ;
			break ;
		case sizeof(short):
			sprintf(buf, "0x%04hx", *((short *) addr)) ;
			break ;
		case sizeof(int):
			sprintf(buf, "0x%08x", *((int *) addr)) ;
			break ;
		case sizeof(long):
			sprintf(buf, "0x%016llx", *((long *) addr)) ;
			break ;
		case (sizeof(long)*2):
			sprintf(buf, "0x%016llx 0x%016llx", *((long *) addr), ((long *) addr)[1]) ;
	}
}


const char **ListToStrArray(Jim_Interp *interp, Jim_Obj *listPtr)
{
	const char **string_array ;
	Jim_Obj *objPtr ;
	int x, y ;
 
	Jim_ListLength(interp, listPtr, &y) ;
	string_array = malloc(sizeof(char *) * (y+1)) ;

	if (string_array == NULL)
		return NULL ;
  
	for (x=0; x<y; x++)
	{
		Jim_ListIndex(interp, listPtr, x, &objPtr, 0) ;
		string_array[x] = Jim_GetString(objPtr, NULL) ;
	}

	string_array[x] = NULL ;
	return string_array ;
}


static int dbgPid(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	char buf[20] ;
	
	sprintf(buf, "%d", dbg->pid) ;
	Jim_SetResultString(interp, buf, -1) ;
	return JIM_OK ;
}
		
		
static int dbgSig(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	const char *options[] = {
		"si_signo", "si_errno", "si_code", "si_pid", "si_uid", "si_status", "si_utime",
		"si_stime", "si_value", "si_int", "si_ptr", "si_addr", "si_band", "si_fd", "send",
		NULL
	} ;
 
	const int offset[] = {
		offsetof(struct Debug, sig.si_signo), offsetof(struct Debug, sig.si_errno),
		offsetof(struct Debug, sig.si_code), offsetof(struct Debug, sig.si_pid),
		offsetof(struct Debug, sig.si_uid), offsetof(struct Debug, sig.si_status),
		offsetof(struct Debug, sig.si_utime), offsetof(struct Debug, sig.si_stime),
		offsetof(struct Debug, sig.si_value), offsetof(struct Debug, sig.si_int),
		offsetof(struct Debug, sig.si_ptr), offsetof(struct Debug, sig.si_addr),
		offsetof(struct Debug, sig.si_band), offsetof(struct Debug, sig.si_fd)
	} ;
 
	const size_t size[] = {
		sizeof(int), sizeof(int), sizeof(int), sizeof(pid_t), sizeof(uid_t),
		sizeof(int), sizeof(clock_t), sizeof(clock_t), sizeof(sigval_t),
		sizeof(int), sizeof(void *), sizeof(void *), sizeof(int), sizeof(int)
	} ;

	long val ;
	int option ;
	char buf[20], *endptr, *dbg = Jim_CmdPrivData(interp) ;
	

	if (argc >= 2)
	{
		if (Jim_GetEnum(interp, argv[1], options, &option, NULL, JIM_ERRMSG) != JIM_OK)
			return JIM_ERR ;

		if (option == 14)
		{
			sprintf(buf,"%d", SendSignal((struct Debug *)dbg)) ;
			Jim_SetResultString(interp, buf, -1) ;
			return JIM_OK ;
		}

		if (argc == 3)
		{
			val = strtoll(Jim_GetString(argv[2], NULL), &endptr, 0) ;
		
			if (endptr == Jim_GetString(argv[2], NULL))
				return JIM_ERR ;

			SetVar(&dbg[offset[option]], &val, size[option]) ;
		}
	
		fVar(&dbg[offset[option]], buf, size[option]) ;
		Jim_SetResultString(interp, buf, -1) ;
		return JIM_OK ;
	}
	
	return JIM_ERR ;
}

		
static int dbgReg(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	const char *options[] = {
	 
		// general
		"eax", "ebx", "ecx", "edx", "esi", "edi", "ebp", "esp", "eip", "orig_eax",
		"rax", "rbx", "rcx", "rdx", "rsi", "rdi", "rbp", "rsp", "rip",
		"r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15", "orig_rax",
		"cs", "ds", "es", "ss", "fs", "gs", "fs_base", "gs_base",

		// fpu
		"cwd", "swd", "twd", "fop", "fip", "fcs",
		"mmx0", "mmx1", "mmx2", "mmx3", "mmx4", "mmx5", "mmx6", "mmx7",
		"fpr0", "fpr1", "fpr2", "fpr3", "fpr4", "fpr5", "fpr6", "fpr7",
	
		// xmm
		"xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7", "xmm8",
		"xmm9", "xmm10", "xmm11", "xmm12", "xmm13", "xmm14", "xmm15", "mxcsr",
	
		// debug
		"dr0", "dr1", "dr2", "dr3", "dr4", "dr5", "dr6", "dr7",
		NULL
	} ;

	const int offset[] = {
		// parte de 32 bits
		offsetof(struct Debug, regs.rax), offsetof(struct Debug, regs.rbx),
		offsetof(struct Debug, regs.rcx), offsetof(struct Debug, regs.rdx),
		offsetof(struct Debug, regs.rsi), offsetof(struct Debug, regs.rdi),
		offsetof(struct Debug, regs.rbp), offsetof(struct Debug, regs.rsp),
		offsetof(struct Debug, regs.rip), offsetof(struct Debug, regs.orig_rax),
		//
		offsetof(struct Debug, regs.rax), offsetof(struct Debug, regs.rbx),
		offsetof(struct Debug, regs.rcx), offsetof(struct Debug, regs.rdx),
		offsetof(struct Debug, regs.rsi), offsetof(struct Debug, regs.rdi),
		offsetof(struct Debug, regs.rbp), offsetof(struct Debug, regs.rsp),
		offsetof(struct Debug, regs.rip), offsetof(struct Debug, regs.r8),
		offsetof(struct Debug, regs.r9), offsetof(struct Debug, regs.r10),
		offsetof(struct Debug, regs.r11), offsetof(struct Debug, regs.r12),
		offsetof(struct Debug, regs.r13), offsetof(struct Debug, regs.r14),
		offsetof(struct Debug, regs.r15), offsetof(struct Debug, regs.orig_rax),
		offsetof(struct Debug, regs.cs), offsetof(struct Debug, regs.ds),
		offsetof(struct Debug, regs.es), offsetof(struct Debug, regs.ss),
		offsetof(struct Debug, regs.fs), offsetof(struct Debug, regs.gs),
		offsetof(struct Debug, regs.fs_base), offsetof(struct Debug, regs.gs_base),
		// fpu/mmx/xmm
#define mmx(x) fregs.st_space[x*4]
#define fpr(x) fregs.st_space[x*4]
		offsetof(struct Debug, fregs.cwd), offsetof(struct Debug, fregs.swd),
		offsetof(struct Debug, fregs.twd), offsetof(struct Debug, fregs.fop),
		offsetof(struct Debug, fregs.rip), offsetof(struct Debug, fregs.rdp),
		offsetof(struct Debug, mmx(0)), offsetof(struct Debug, mmx(1)),
		offsetof(struct Debug, mmx(2)), offsetof(struct Debug, mmx(3)),
		offsetof(struct Debug, mmx(4)), offsetof(struct Debug, mmx(5)),
		offsetof(struct Debug, mmx(6)), offsetof(struct Debug, mmx(7)),
		offsetof(struct Debug, fpr(0)), offsetof(struct Debug, fpr(1)),
		offsetof(struct Debug, fpr(2)), offsetof(struct Debug, fpr(3)),
		offsetof(struct Debug, fpr(4)), offsetof(struct Debug, fpr(5)),
		offsetof(struct Debug, fpr(6)), offsetof(struct Debug, fpr(7)),
#define xmm(x) fregs.xmm_space[x*4]
		offsetof(struct Debug, xmm(0)), offsetof(struct Debug, xmm(1)),
		offsetof(struct Debug, xmm(2)), offsetof(struct Debug, xmm(3)),
		offsetof(struct Debug, xmm(4)), offsetof(struct Debug, xmm(5)),
		offsetof(struct Debug, xmm(6)), offsetof(struct Debug, xmm(7)),
		offsetof(struct Debug, xmm(8)), offsetof(struct Debug, xmm(9)),
		offsetof(struct Debug, xmm(10)), offsetof(struct Debug, xmm(11)),
		offsetof(struct Debug, xmm(12)), offsetof(struct Debug, xmm(13)),
		offsetof(struct Debug, xmm(14)), offsetof(struct Debug, xmm(15)),
		offsetof(struct Debug, fregs.mxcsr),
		// debug
		offsetof(struct Debug, u_debugreg[0]), offsetof(struct Debug, u_debugreg[1]),
		offsetof(struct Debug, u_debugreg[2]), offsetof(struct Debug, u_debugreg[3]),
		offsetof(struct Debug, u_debugreg[4]), offsetof(struct Debug, u_debugreg[5]),
		offsetof(struct Debug, u_debugreg[6]), offsetof(struct Debug, u_debugreg[7]),
	} ;
	
	const size_t size[] = {
		sizeof(int), sizeof(int), sizeof(int), sizeof(int), sizeof(int), sizeof(int),
		sizeof(int), sizeof(int), sizeof(int), sizeof(int), sizeof(long), sizeof(long),
		sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(long),
		sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(long),
		sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(short), sizeof(short),
		sizeof(short), sizeof(short), sizeof(short), sizeof(short), sizeof(long),
		sizeof(long), sizeof(short), sizeof(short), sizeof(short), sizeof(short),
		sizeof(long), sizeof(long), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2), (sizeof(long)*2),
		(sizeof(long)*2), (sizeof(long)*2), sizeof(long), sizeof(long), sizeof(long),
		sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(long), sizeof(long)
	} ;

	long val[2] = { 0, 0 } ;
	int option ;
	char buf[20], *endptr, *dbg = Jim_CmdPrivData(interp) ;

	
	if (argc >= 2)
	{
		if (Jim_GetEnum(interp, argv[1], options, &option, NULL, JIM_ERRMSG) != JIM_OK)
		return JIM_ERR ;

		if (argc == 3)
		{
			if (size[option] == (sizeof(long)*2))
				if(sscanf(Jim_GetString(argv[2], NULL), "%llx %llx",&val[0],&val[1]) == 1)
				{
					val[1] = val[0] ;
					val[0] = 0 ;
				}
			else
				*((long *) &val) = strtoll(Jim_GetString(argv[2], NULL), &endptr, 0) ;
		
			if (endptr == Jim_GetString(argv[2], NULL))
				return JIM_ERR ;

			SetVar(&dbg[offset[option]], &val, size[option]) ;
		}
	
		fVar(&dbg[offset[option]], buf, size[option]) ;
		Jim_SetResultString(interp, buf, -1) ;
		return JIM_OK ;
	}
	
	return JIM_ERR ;
}


static int dbgMap(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	struct vm_map map ;
	unsigned long addr ;
	char buf[PATH_MAX+25+sizeof(void*)*6], *endptr ;

	if (argc != 2)
		return JIM_ERR ;

	addr = strtoll(Jim_GetString(argv[1], NULL), &endptr, 0) ;

	if (endptr == Jim_GetString(argv[1], NULL))
		return JIM_ERR ;

	buf[0] = '\0' ;
  
	if (GetMapInfo(addr, &map, dbg))
		goto out ;

	sprintf(
		buf,
		"0x%016llx 0x%016llx %s %s %s %s 0x%016llx %02x %02x %s",
		map.start,
		map.end,
		(map.flags & VM_READ) ? "r":"{}",
		(map.flags & VM_WRITE) ? "w":"{}",
		(map.flags & VM_EXEC) ? "x":"{}",
		(map.flags & VM_SHARED) ? "s":"p",
		map.offset,
		map.dev,
		map.inode,
		map.file
		) ;
out:
	Jim_SetResultString(interp, buf, -1) ;
	return JIM_OK ;
}


static inline void dbgDelCommands(Jim_Interp *interp, const char *cmd)
{
	unsigned int a ;
	char buf[128] ;
	
	strcpy(buf,cmd) ;
	
	for (a = 0; a < CmdIndex; a++)
	{
		(strchr(buf, '.'))[1] = '\0' ;
		strcat(buf, Commands[a].name) ;
		Jim_DeleteCommand(interp, buf) ;
	}
}


static int dbgKill(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	int r = Kill(dbg) ;
	char buf[20] ;
		 
	/* falle o no limpiamos todo (el proceso ya puede haber muerto) */
	RemoveAllBp(dbg) ;
	RemoveAllWp(dbg) ;
	dbgDelCommands(interp, Jim_GetString(argv[0], NULL)) ;
	Jim_Free(dbg) ;

	sprintf(buf, "%d", r) ;
	Jim_SetResultString(interp, buf, -1) ;
	return JIM_OK ;
}


static int dbgDetach(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	int r = Detach(dbg) ;
	char buf[20] ;
  
	if (!r)
	{
		RemoveAllBp(dbg) ;
		RemoveAllWp(dbg) ;
		dbgDelCommands(interp, Jim_GetString(argv[0], NULL)) ;
		Jim_Free(dbg) ;
	}
 
	sprintf(buf, "%d", r) ;
	Jim_SetResultString(interp, buf, -1) ;
	return JIM_OK ;
}


#define dbgMake(x) static int dbg##x(Jim_Interp *interp, int argc, Jim_Obj *const *argv) \
{ \
	struct Debug *dbg = Jim_CmdPrivData(interp) ; \
	char buf[20] ; \
	               \
	SetRegs(dbg) ; \
	sprintf(buf, "%d", x(dbg)) ; \
	Jim_SetResultString(interp, buf, -1) ; \
	return JIM_OK ; \
}

dbgMake(Run)
dbgMake(StepIn)
dbgMake(StepOver)
dbgMake(Syscall)


static int dbgBp(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	const char *options[] = {
		 "opcode", "stat", "list", "set", "unset", "fix", "fixall", "unsetall", NULL
	} ;
	
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	struct BreakPoint *bps ;
	unsigned long addr = -1 ;
	int option ;
	char *buf, *endptr ;

	buf = malloc(460) ;
	buf[0] = '\0' ;

	if (argc >= 2)
	{
		if (Jim_GetEnum(interp, argv[1], options, &option, NULL, JIM_ERRMSG) != JIM_OK)
			goto out ;

		if (argc == 3)
		{
			addr = strtoll(Jim_GetString(argv[2], NULL), &endptr, 0) ;
					
			if (endptr == Jim_GetString(argv[2], NULL))
				goto out ;
		}

		switch(option)
		{
			case 0:
				if (argc == 3)
					BpOpcode = (char) addr ;
				
				sprintf(buf, "%d", BpOpcode) ;
				break ;
			case 1:
				
				if (!(bps = FindBp(addr, dbg)))
					break ;

				sprintf(buf, "0x%016llx 0x%02hhx 0x%02hhx", bps->addr, bps->op, bps->bp) ;
				break ;
			case 2:
				bps = dbg->bps ;
				addr = 0 ;

				while(bps)
				{
					sprintf(&buf[addr], "{0x%016llx %c} ", bps->addr, (StillBpSet(bps, dbg))?'1':'0') ;
					addr += 23 ;

					if (addr >= malloc_usable_size(buf))
						buf = realloc(buf, malloc_usable_size(buf) + 460) ;
									       
					bps = bps->next ;
				}

				if (addr)
					addr-- ;
				
				buf[addr] = '\0' ;
				break ;
			case 3:
				if (argc == 3)
				{
					sprintf(buf, "%d", SetBp(addr, dbg)) ;
					break ;
				}
				
				goto out ;
			case 4:
				if (argc == 3)
				{
					sprintf(buf, "%d", RemoveBp(addr, dbg)) ;
					break ;
				}
				
				goto out ;
			case 5:
				if (argc == 3)
				{
					sprintf(buf, "%d", ResetBp(addr, dbg)) ;
					break ;
				}
				
				goto out ;
			case 6:
				sprintf(buf, "%d", ResetAllBp(dbg)) ;
				break ;
			case 7:
				sprintf(buf, "%d", RemoveAllBp(dbg)) ;
		}

		Jim_SetResultString(interp, buf, -1) ;
		free(buf) ;
		return JIM_OK ;
	}
out:
	free(buf) ;
	return JIM_ERR ;
}



static int dbgWp(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	const char *options[] = {
		 "stat", "list", "set", "unset", "fix", "fixall", "unsetall", NULL
	} ;
	
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	struct WatchPoint *wps ;
	struct WpInfo wpi ;
	unsigned long addr = -1, size, flags ;
	int option ;
	char *buf, *endptr ;

	buf = malloc(840) ;
	buf[0] = '\0' ;

	if (argc >= 2)
	{
		if (Jim_GetEnum(interp, argv[1], options, &option, NULL, JIM_ERRMSG) != JIM_OK)
			goto out ;

		switch(argc)
		{
			case 5:
				flags = strtoll(Jim_GetString(argv[4], NULL), &endptr, 0) ;
					
				if (endptr == Jim_GetString(argv[4], NULL))
					goto out ;
			case 4:
				size = strtoll(Jim_GetString(argv[3], NULL), &endptr, 0) ;
					
				if (endptr == Jim_GetString(argv[3], NULL))
					goto out ;
			case 3:
				addr = strtoll(Jim_GetString(argv[2], NULL), &endptr, 0) ;
					
				if (endptr == Jim_GetString(argv[2], NULL))
					goto out ;
		}

		switch(option)
		{
			case 0:
				if (argc != 3)
					goto out ;
				
				wpi.addr = addr ;
				
				if (GetWpInfo(&wpi, dbg) == -1)
					break ;
				
				sprintf(
					buf,
					"0x%016llx 0x%016llx %s %s %s %s",
					wpi.addr,
					wpi.end,
					(wpi.flags & WP_READ) ? "r":"{}",
					(wpi.flags & WP_WRITE) ? "w":"{}",
					(wpi.flags & WP_EXEC) ? "x":"{}",
					(wpi.flags & WP_ACCESS) ? "a":"{}"
				       ) ;
				
				break ;
			case 1:
				wps = dbg->wps ;
				addr = 0 ;

				while(wps)
				{
					wpi.addr = wps->addr ;
				head:
					while(GetWpInfo(&wpi, dbg) == -1)
					{
						wpi.addr++ ;

						if (wpi.addr == (wps->addr + PAGE_SIZE))
							goto tail ;
					}
					 
					sprintf(
						&buf[addr], "{0x%016llx 0x%016llx %c} ",
						wpi.addr, wpi.end, (StillWpSet(wps, dbg))?'1':'0'
					       ) ;
					
					addr += 42 ;
					
					if (addr >= malloc_usable_size(buf))
						buf = realloc(buf, malloc_usable_size(buf) + 840) ;

					wpi.addr = wpi.end ;
					goto head ;
					
				tail:
					wps = wps->next ;
				}

				if (addr)
					addr-- ;
				
				buf[addr] = '\0' ;
				break ;
			case 2:
				if (argc == 5)
				{
					sprintf(buf, "0x%016llx", SetWp(addr, size, flags, dbg)) ;
					break ;
				}
				
				goto out ;
			case 3:
				if (argc == 4)
				{
					sprintf(buf, "%d", RemoveWp(addr, size, dbg)) ;
					break ;
				}
				
				goto out ;
			case 4:
				if (argc == 3)
				{
					sprintf(buf, "%d", ResetWp(addr, dbg)) ;
					break ;
				}
				
				goto out ;
			case 5:
				sprintf(buf, "%d", ResetAllWp(dbg)) ;
				break ;
			case 6:
				sprintf(buf, "%d", RemoveAllWp(dbg)) ;
		}

		Jim_SetResultString(interp, buf, -1) ;
		free(buf) ;
		return JIM_OK ;
	}
out:
	free(buf) ;
	return JIM_ERR ;
}


static int dbgRead(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	unsigned long addr, size ;
	char *buf, *endptr ;
	int x ;
	
	if (argc != 3)
	{
		Jim_WrongNumArgs(interp, 1, argv, "addr size") ;
		return JIM_ERR ;
	}

	buf = (char *) Jim_GetString(argv[1], NULL) ;
	addr = strtoll(buf, &endptr, 0) ;
	
	if (endptr == buf)
		return JIM_ERR ;

	buf = (char *) Jim_GetString(argv[2], NULL) ;
	size = strtoll(buf, &endptr, 0) ;

	if (endptr == buf)
		return JIM_ERR ;

	buf = malloc(size) ;
	size = ReadMem(addr, size, buf, dbg) ;
	endptr = malloc(size*5) ;
	
	for(x=0;x < size;x++)
		sprintf(&endptr[x*5], "0x%02hhx ",buf[x]) ;

	free(buf) ;
	
	Jim_SetResultString(interp, endptr, -1) ;
	free(endptr) ;
	return JIM_OK ;
}


static int dbgRread(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	unsigned long addr, size ;
	char *buf, *endptr ;
	int x ;
	
	if (argc != 3)
	{
		Jim_WrongNumArgs(interp, 1, argv, "addr size") ;
		return JIM_ERR ;
	}

	buf = (char *) Jim_GetString(argv[1], NULL) ;
	addr = strtoll(buf, &endptr, 0) ;
	
	if (endptr == buf)
		return JIM_ERR ;

	buf = (char *) Jim_GetString(argv[2], NULL) ;
	size = strtoll(buf, &endptr, 0) ;

	if (endptr == buf)
		return JIM_ERR ;

	buf = malloc(size) ;
	size = RReadMem(addr, size, buf, dbg) ;
	endptr = malloc(size*5) ;
	
	for(x=0;x < size;x++)
		sprintf(&endptr[x*5], "0x%02hhx ",buf[x]) ;

	free(buf) ;
	
	Jim_SetResultString(interp, endptr, -1) ;
	free(endptr) ;
	return JIM_OK ;
}


static int dbgWrite(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg = Jim_CmdPrivData(interp) ;
	unsigned long addr, size ;
	char *buf, *buf2, *endptr ;
	int x ;
	
	if (argc != 3)
	{
		Jim_WrongNumArgs(interp, 1, argv, "addr data") ;
		return JIM_ERR ;
	}

	buf = (char *) Jim_GetString(argv[1], NULL) ;
	addr = strtoll(buf, &endptr, 0) ;
	
	if (endptr == buf)
		return JIM_ERR ;

	buf = (char *) Jim_GetString(argv[2], NULL) ;
	buf2 = malloc(strlen(buf)) ;

	for(x=0; *buf != '\0'; x++)
	{
		buf2[x] = (char) strtol(buf,&endptr,0) ;
		
		if (buf == endptr)
		{
			free(buf2) ;
			return JIM_ERR ;
		}
		buf = endptr ;
	}
	
	sprintf(buf2,"%d",WriteMem(addr, x, buf2, dbg)) ;
	Jim_SetResultString(interp, buf2, -1) ;
	free(buf2) ;
	return JIM_OK ;
}


static int dbgTrace(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	const char **_argv = NULL ;
	const char **_envp = NULL ;
	struct Debug *dbg = Jim_Alloc(sizeof(struct Debug)) ;
	Jim_Obj *objPtr ;
	unsigned int a ;
	char buf[128] ;
 
	if (!dbg)
		return JIM_ERR ;

	memset(dbg, 0, sizeof(struct Debug)) ;
  
	switch(argc)
	{
		case 4:
			if ((_envp = ListToStrArray(interp, argv[3])) == NULL)
				break ;
		case 3:
			if ((_argv = ListToStrArray(interp, argv[2])) == NULL)
				break ;
		case 2:
			if (TraceFile(Jim_GetString(argv[1], NULL), (char *const *)_argv,(char *const *)_envp, dbg))
			{
				buf[0] = '\0' ;
				Jim_Free(dbg) ;
				goto end ;
			}
		
			sprintf(buf, "set dbg_count %lu", dbgIndex) ;
			Jim_EvalGlobal(interp, buf) ;
			sprintf(buf, "dbg%lu.", dbgIndex) ;
		
			for (a = 0; a < CmdIndex; a++)
			{
				(strchr(buf, '.'))[1] = '\0' ;
				strcat(buf, Commands[a].name) ;
				Jim_CreateCommand(interp, buf, Commands[a].dbgCommand, dbg, NULL) ;
			}
		
			dbgIndex++ ;
			(strchr(buf,'.'))[0] = '\0' ;
		end:
				Jim_SetResultString(interp, buf, -1) ;
		free(_envp) ;
		free(_argv) ;
		return JIM_OK ;

		default:
			Jim_WrongNumArgs(interp, 1, argv, "filename argv envp") ;
	}
	
	free(_envp) ;
	free(_argv) ;
	Jim_Free(dbg) ;
	return JIM_ERR ;
}


static int dbgAttach(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
{
	struct Debug *dbg ;
	unsigned int a ;
	pid_t pid ;
	char buf[20], *endptr ;
  
	if (argc != 2)
		return JIM_ERR ;

	pid = strtoll(Jim_GetString(argv[1], NULL), &endptr, 0) ;

	if (endptr == Jim_GetString(argv[1], NULL))
		return JIM_ERR ;
  
	if (!(dbg = Jim_Alloc(sizeof(struct Debug))))
	{
		buf[0] = '\0' ;
		goto out ;
	}

	memset(dbg, 0, sizeof(struct Debug)) ;
	dbg->pid = pid ;

	if (Attach(dbg))
	{
		buf[0] = '\0' ;
		Jim_Free(dbg) ;
		goto out ;
	}
	
	sprintf(buf, "set dbg_count %lu", dbgIndex) ;
	Jim_EvalGlobal(interp, buf) ;
	sprintf(buf, "dbg%lu.", dbgIndex) ;
  
	for (a = 0; a < CmdIndex; a++)
	{
		(strchr(buf, '.'))[1] = '\0' ;
		strcat(buf, Commands[a].name) ;
		Jim_CreateCommand(interp, buf, Commands[a].dbgCommand, dbg, NULL) ;
	}
		
	dbgIndex++ ;
	(strchr(buf,'.'))[0] = '\0' ;
out:
	Jim_SetResultString(interp, buf, -1) ;
	return JIM_OK ;
}


int main()
{
	Jim_Interp *interp ;

	dbgAddCommand("bp", dbgBp) ;
	dbgAddCommand("wp", dbgWp) ;
	dbgAddCommand("pid", dbgPid) ;
	dbgAddCommand("sig", dbgSig) ;
	dbgAddCommand("reg", dbgReg) ;
	dbgAddCommand("map", dbgMap) ;
	dbgAddCommand("run", dbgRun) ;
	dbgAddCommand("kill", dbgKill) ;
	dbgAddCommand("step", dbgStepIn) ;
	dbgAddCommand("read", dbgRead) ;
	dbgAddCommand("rread", dbgRread) ;
	dbgAddCommand("write", dbgWrite) ;
	dbgAddCommand("detach", dbgDetach) ;
	dbgAddCommand("syscall", dbgSyscall) ;
	dbgAddCommand("step_over", dbgStepOver) ;

	Jim_InitEmbedded() ;
	interp = Jim_CreateInterp() ;
	Jim_RegisterCoreCommands(interp) ;

	Jim_CreateCommand(interp, "trace", dbgTrace, NULL, NULL) ;
	Jim_CreateCommand(interp, "attach", dbgAttach, NULL, NULL) ;

	if (Jim_EvalFile(interp, "patan.tcl") != JIM_OK)
		Jim_PrintErrorMessage(interp) ;

	Jim_SetVariableStrWithStr(interp, "jim_interactive", "1") ;
	Jim_InteractivePrompt(interp) ;
	Jim_FreeInterp(interp) ;
	return 0 ;
}


